from .support import TickLiveData, MinLiveData, TouchlineData
from .support import TDLiveDataError, TDHistoricDataError
from .TD_live import LiveClient
from .TD_hist import HistoricalREST
from .TD_chain import OptionChain
from threading import Thread
from typing import Union, Callable, Any
from collections import defaultdict

from datetime import datetime
from dateutil.relativedelta import relativedelta
from typing import List, Dict
from colorama import Style, Fore
import time
import json
import logging

# import traceback

class TD:
    def __init__(self,
                 login_id,
                 password,
                 broker_token=None,
                 url='push.truedata.in',  # This arg is used for LIVE URL only
                 live_port=8082,
                 historical_api=True,
                 # tz='Asia/Kolkata',
                 log_level=logging.WARNING,
                 log_handler=None,
                 log_format=None,
                 hist_url='https://history.truedata.in'):
        self.live_websocket = None
        self.historical_datasource = None
        self.login_id = login_id
        self.password = password
        self.live_url = url
        self.hist_url = hist_url
        self.live_port = live_port
        self.connect_historical = historical_api
        if log_format is None:
            # log_format = "%(levelname)s : %(message)s"
            log_format = "(%(asctime)s) %(levelname)s :: %(message)s (PID:%(process)d Thread:%(thread)d)"
        if log_handler is None:
            log_formatter = logging.Formatter(log_format)
            # log_formatter = logging.Formatter("%(message)s")
            self.log_handler = logging.StreamHandler()
            self.log_handler.setLevel(log_level)
            self.log_handler.setFormatter(log_formatter)
        else:
            self.log_handler = log_handler
        self.logger = logging.getLogger(__name__)
        self.logger.addHandler(self.log_handler)
        self.logger.setLevel(self.log_handler.level)
        self.logger.debug("Logger ready...")
        if live_port is None:
            self.connect_live = False
        else:
            self.connect_live = True
        self.broker_token = broker_token
        self.hist_data = {}
        self.live_data = {}
        self.min_live_data = {}
        self.symbol_mkt_id_map = defaultdict(set)
        # self.streaming_symbols = {}
        self.touchline_data = {}
        self.default_market_data_id = 2000
        self.connect()

    def connect(self):
        broker_append = ''
        if self.broker_token is not None:
            broker_append = f'&brokertoken={self.broker_token}'
        if self.connect_live:
            self.live_websocket = LiveClient(self, f"wss://{self.live_url}:{self.live_port}?user={self.login_id}&password={self.password}{broker_append}")
            t = Thread(target=self.connect_thread, args=(), daemon=True)
            t.start()
        if self.connect_historical:
            self.historical_datasource = HistoricalREST(self.login_id, self.password, self.hist_url, self.broker_token, self.logger)
        while self.connect_live and self.live_websocket.subscription_type == '':
            time.sleep(1)

    def connect_thread(self):
        self.live_websocket.run_forever(ping_interval=10, ping_timeout=5)
        while not self.live_websocket.disconnect_flag:
            self.logger.info(f"{Style.BRIGHT}{Fore.RED}Connection dropped due to no network , Attempting reconnect @ {Fore.CYAN}{datetime.now()}{Style.RESET_ALL}")
            self.live_websocket.reconnect()
            time.sleep(10)
        self.logger.debug('Goodbye (properly) !!')

    def disconnect(self):
        if self.connect_live:
            self.live_websocket.disconnect_flag = True
            self.live_websocket.close()
            # self.logger.info(f"{Fore.GREEN}Disconnected LIVE TrueData...{Style.RESET_ALL}")
            self.logger.warning(f"{Style.NORMAL}{Fore.BLUE}Disconnected from Real Time Data WebSocket Connection !{Style.RESET_ALL}")
        if self.connect_historical:
            # self.logger.info(f"{Fore.GREEN}Disconnected HISTORICAL TrueData...{Style.RESET_ALL}")
            self.logger.warning(f"{Style.NORMAL}{Fore.BLUE}Disconnected from Historical Data REST Connection !{Style.RESET_ALL}")

    @staticmethod
    def truedata_duration_map(regular_format, end_date):
        duration_units = regular_format.split()[1].upper()
        if len(duration_units) > 1:
            raise TDHistoricDataError("Misconfigured duration argument")
        duration_size = int(regular_format.split()[0])
        if duration_units == 'D':
            return (end_date - relativedelta(days=duration_size - 1)).date()
        elif duration_units == 'W':
            return (end_date - relativedelta(weeks=duration_size)).date()
        elif duration_units == 'M':
            return (end_date - relativedelta(months=duration_size)).date()
        elif duration_units == 'Y':
            return (end_date - relativedelta(years=duration_size)).date()

    def get_historic_data(self, contract,
                          ticker_id=None,
                          end_time=None,
                          duration=None,
                          start_time=None,
                          bar_size="1 min",
                          options=None,
                          bidask=False):
        if start_time is not None and duration is None:
            return self.get_historical_data_from_start_time(contract=contract,
                                                            ticker_id=ticker_id,
                                                            end_time=end_time,
                                                            start_time=start_time,
                                                            bar_size=bar_size,
                                                            options=options,
                                                            bidask=bidask)
        else:
            return self.get_historical_data_from_duration(contract=contract,
                                                          ticker_id=ticker_id,
                                                          end_time=end_time,
                                                          duration=duration,
                                                          bar_size=bar_size,
                                                          options=options,
                                                          bidask=bidask)

    def get_n_historical_bars(self, contract,
                              ticker_id=None,
                              end_time: datetime = None,
                              no_of_bars: int = 1,
                              bar_size="1 min",
                              options=None,
                              bidask=False):
        if end_time is None:
            end_time = datetime.today()
        end_time = end_time.strftime('%y%m%dT%H:%M:%S')    # This is the request format
        hist_data = self.historical_datasource.get_n_historic_bars(contract,
                                                                   end_time,
                                                                   no_of_bars,
                                                                   bar_size,
                                                                   options=options,
                                                                   bidask=bidask)
        if ticker_id is not None:
            self.hist_data[ticker_id] = hist_data
        return hist_data

    def get_historical_data_from_duration(self, contract,
                                          ticker_id=None,
                                          end_time: datetime = None,
                                          duration=None,
                                          bar_size="1 min",
                                          options=None,
                                          bidask=False):
        if duration is None:
            duration = "1 D"
        if end_time is None:
            end_time = datetime.today()
        start_time = self.truedata_duration_map(duration, end_time)
        end_time = end_time.strftime('%y%m%d') + 'T23:59:59'    # This is the request format
        start_time = start_time.strftime('%y%m%d') + 'T00:00:00'    # This is the request format

        hist_data = self.historical_datasource.get_historic_data(contract, end_time, start_time, bar_size, options=options, bidask=bidask)
        if ticker_id is not None:
            self.hist_data[ticker_id] = hist_data
        return hist_data

    def get_historical_data_from_start_time(self, contract,
                                            ticker_id=None,
                                            end_time: datetime = None,
                                            start_time: datetime = None,
                                            bar_size="1 min",
                                            options=None,
                                            bidask=False):
        if end_time is None:
            end_time = datetime.today().replace(hour=23, minute=59, second=59)
        if start_time is None:
            start_time = datetime.today().replace(hour=0, minute=0, second=0)

        end_time = end_time.strftime('%y%m%dT%H:%M:%S')    # This is the request format
        start_time = start_time.strftime('%y%m%dT%H:%M:%S')    # This is the request format
        hist_data = self.historical_datasource.get_historic_data(contract, end_time, start_time, bar_size, options=options, bidask=bidask)
        if ticker_id is not None:
            self.hist_data[ticker_id] = hist_data
        return hist_data

    def get_bhavcopy(self, segment: str, date: datetime = None, return_completed: bool = True) -> List[Dict]:
        if date is None:
            date = datetime.now().replace(hour=0, minute=0, second=0)
        # return_completed = True  # Uncomment and remove arg to disallow user choice
        return self.historical_datasource.bhavcopy(segment, date, return_completed)

    def get_gainers(self , segment , topn , df_style = True ):
        topn = 10 if topn is None else topn
        return self.historical_datasource.get_gainers_losers(segment, topn , gainers = True , df_style = df_style )

    def get_losers(self , segment , topn , df_style = True ):
        topn = 10 if topn is None else topn
        return self.historical_datasource.get_gainers_losers(segment , topn , gainers = False , df_style = df_style )


    @staticmethod
    def get_req_id_list(req_id: Union[int, list], len_contracts: int) -> list:
        if type(req_id) == list:
            if len(req_id) == len_contracts:
                return req_id
            else:
                raise TDLiveDataError("Lengths do not match...")
        elif type(req_id) == int:
            curr_req_id = req_id
            return [curr_req_id + i for i in range(0, len_contracts)]
        else:
            raise TDLiveDataError("Invalid req_id datatype...")

    def start_live_data(self, resolved_contracts, req_id=None):  # TODO: Prevent reuse of req_ids
        symbols_to_call = []
        if req_id is not None:
            req_ids = self.get_req_id_list(req_id, len(resolved_contracts))
        else:
            req_ids = self.get_req_id_list(self.default_market_data_id, len(resolved_contracts))
            self.default_market_data_id = self.default_market_data_id + len(resolved_contracts)
        for i , req_id in enumerate(req_ids):
            resolved_contract = resolved_contracts[i].upper()
            if resolved_contract in self.symbol_mkt_id_map.keys():
                self.touchline_data[req_id] = self.touchline_data[list(self.symbol_mkt_id_map[resolved_contract])[0]]
            else:
                self.touchline_data[req_id] = TouchlineData()
            if self.live_websocket.subscription_type == 'tick':
                if resolved_contract in self.symbol_mkt_id_map.keys():
                    self.live_data[req_id] = self.live_data[list(self.symbol_mkt_id_map[resolved_contract])[0]]
                elif req_id not in self.live_data.keys() or type(self.live_data[req_id]) != TickLiveData:
                    self.live_data[req_id] = TickLiveData(resolved_contract)
            elif '+' in self.live_websocket.subscription_type:
                if resolved_contract in self.symbol_mkt_id_map.keys():
                    self.live_data[req_id] = self.live_data[list(self.symbol_mkt_id_map[resolved_contract])[0]]
                elif req_id not in self.live_data.keys() or type(self.live_data[req_id]) != TickLiveData:
                    self.live_data[req_id] = TickLiveData(resolved_contract)
                if resolved_contract in self.symbol_mkt_id_map.keys():
                    self.min_live_data[req_id] = self.min_live_data[list(self.symbol_mkt_id_map[resolved_contract])[0]]
                elif req_id not in self.min_live_data.keys() or type(self.min_live_data[req_id]) != MinLiveData:
                    self.min_live_data[req_id] = MinLiveData(resolved_contract)
            elif 'min' in self.live_websocket.subscription_type:
                if resolved_contract in self.symbol_mkt_id_map.keys():
                    self.live_data[req_id] = self.live_data[list(self.symbol_mkt_id_map[resolved_contract])[0]]
                elif req_id not in self.live_data.keys() or type(self.live_data[req_id]) != MinLiveData:
                    self.live_data[req_id] = MinLiveData(resolved_contract)
            if resolved_contract not in self.symbol_mkt_id_map.keys():
                symbols_to_call.append(resolved_contract)
            self.symbol_mkt_id_map[resolved_contract].update({req_id})
        if len(symbols_to_call) > 0:
            self.live_websocket.send(f'{{"method": "addsymbol", "symbols": {json.dumps(symbols_to_call)}}}')
        return req_ids

    def stop_live_data(self, contracts):  # Clearing objects is done after server confirmation
        remove_symbols = []
        for sym in contracts:
            if len(self.symbol_mkt_id_map[sym]) > 1:
                self.symbol_mkt_id_map[sym].remove(list(self.symbol_mkt_id_map[sym])[-1])
            else:
                remove_symbols.append(sym)
                self.symbol_mkt_id_map.pop(sym , None)
        self.live_websocket.send(f'{{"method": "removesymbol", "symbols": {json.dumps(remove_symbols)}}}')

    def trade_callback(self, func: Callable[[TickLiveData], Any]):
        self.logger.info(f"Defining {func} as trade_callback...")
        self.live_websocket.trade_callback = func

    def clear_trade_callback(self):
        self.logger.info(f"Clearing trade_callback...")
        self.live_websocket.trade_callback = None

    def bidask_callback(self, func: Callable[[TickLiveData], Any]):
        self.logger.info(f"Defining bidask_callback...")
        self.live_websocket.bidask_callback = func

    def clear_bidask_callback(self):
        self.logger.info(f"Clearing bidask_callback...")
        self.live_websocket.bidask_callback = None

    def bar_callback(self, func: Callable[[TickLiveData], Any]):
        self.logger.info(f"Defining bar_callback...")
        self.live_websocket.bar_callback = func

    def clear_bar_callback(self):
        self.logger.info(f"Clearing bar_callback...")
        self.live_websocket.bar_callback = None

    # def get_touchline(self):
    #     self.live_websocket.send(f'{{"method": "touchline"}}')

    def start_option_chain( self , symbol , expiry , chain_length = None , bid_ask = False , market_open_post_hours = False ):
        chain_length = 10 if chain_length is None else chain_length  # setting default value for chain length
        future_price = self.get_n_historical_bars(f'{symbol}-I' )[0]['c']
        try:
            chain = OptionChain(self , symbol, expiry , chain_length , future_price , bid_ask  , market_open_post_hours )
        except Exception as e:
            # traceback.print_exc()
            print(f'Please check symbol: {symbol} and its expiry: {expiry.date()}')
            exit()
        symb_ids = self.start_live_data(chain.option_symbols)
        option_thread = Thread(target=chain.update_chain, args=( self , symb_ids ), daemon=True)
        option_thread.start()
        return chain



from . import websocket
from .websocket.defaults import MIN_PYTHON_MINOR, MIN_PYTHON_MAJOR
from colorama import init  # , Style, Fore, Back
# import sys
# import subprocess
# from time import sleep
# from copy import deepcopy
# from distutils import version
# import logging

init()  # Colorama

__version__ = '4.2.4'
# logger = logging.getLogger(__name__)
# logger.propagate = True

# try:
#     user_python_version = sys.version_info
#     print(user_python_version)
#     if not (user_python_version.major >= MIN_PYTHON_MAJOR and user_python_version.minor >= MIN_PYTHON_MINOR):
#         logger.warning(f"{Style.BRIGHT}{Fore.RED}\tIt is highly advised that you use Python {MIN_PYTHON_MAJOR}.{MIN_PYTHON_MINOR} or greater...\n"
#                        f"\tSome features may not work as intended on your version ({user_python_version.major}.{user_python_version.minor}.{user_python_version.micro})...{Style.RESET_ALL}")
#         logger.warning(f"\tPlease decide if you want to proceed with running as is or not...\n")
#         wrong_version_warning_input = input(f"\tReply with: {Back.CYAN}{Fore.BLACK}Y/N{Style.RESET_ALL} (Anything other than {Back.CYAN}{Fore.BLACK}Y{Style.RESET_ALL} will exit the code) : ")
#         if wrong_version_warning_input.upper()[0] != "Y":
#             exit()
# except Exception as e:
#     logger.error(f'{Style.BRIGHT}{Fore.RED}\tUnable to get the version of Python you are using...\n'
#                  f'\tPlease report this error to truedata...\n'
#                  f'\tAlong with the following information - {e}...{Style.RESET_ALL}')
# try:
#     op = subprocess.run([sys.executable, '-m', 'pip', 'install', '--upgrade', 'truedata_ws=='], capture_output=True)
#     latest_version = max([version.StrictVersion(i.strip()) for i in str(op.stderr).split('from versions:')[1].split(')')[0].split(',')])
#     if latest_version > version.StrictVersion(__version__):
#         if version.StrictVersion(__version__) < version.StrictVersion('0.1.0'):
#             logger.warning(f'{Style.BRIGHT}{Fore.GREEN}Using DEVELOPER version (deVer-{__version__})...{Style.RESET_ALL}')
#         else:
#             logger.warning(f"{Style.BRIGHT}{Fore.GREEN}\tThere is a newer version of this library available ({latest_version}), while your version is {__version__}...\n"
#                            f"\tIt is highly advisable that you keep your libraries up-to-date... Please upgrade your library using-\n"
#                            f"\n\t\t python3.7 -m pip install --upgrade truedata_ws\n"
#                            f"\tWe also strongly recommend you read our release notes at the end of the README.md that is found at the end of PyPi page "
#                            f"(found at: https://pypi.org/project/truedata-ws/){Style.RESET_ALL}")
# except Exception as e:
#     # logger.error(f'{Style.BRIGHT}{Fore.RED}\tUnable to get latest version from PyPi...\n'
#     #              f'\tPlease report this error to truedata (https://github.com/kapilmar/truedata_ws/issues/new)...\n'
#     #              f'\tAlong with the following information - {e}...{Style.RESET_ALL}')
#     pass



import sys
from .websocket import client_tests, internal_tests
from . import __version__

if sys.argv[1] == 'version':
    print(__version__)
    sys.exit(0)

if sys.argv[1] == 'run_all_tests':
    print(f'Running all tests on version = {__version__}')
    username = sys.argv[2]
    password = sys.argv[3]
    live_port = None
    hist_port = None
    symbols = None
    try:
        for arg in sys.argv[4:]:
            if arg[:2] == "--":
                arg_split = arg[2:].split('=')
                arg_key = arg_split[0]
                arg_value = arg_split[1]
                if arg_key.lower() == 'liveport':
                    live_port = int(arg_value)
                if arg_key.lower() == 'histport':
                    hist_port = int(arg_value)
                if arg_key.lower() == 'symbols':
                    symbols = arg_value.split()
                    symbols = [symbol.strip() for symbol in symbols]
    except IndexError:
        pass
    client_tests.run_all_tests(username, password, live_port, hist_port, symbols)


if sys.argv[1] == 'run_internal_tests':
    print(f'The setup:\n'
          f'\tAccount A with live data = tick streaming, and bid-ask enabled in historic data.\n'
          f'\tAccount B with live data = 1min streaming, and no bid-ask in history.\n'
          f'\tThe live stream will run for 5 mins before terminating...')
    print(f'Running all tests on version = {__version__}')
    user_A = input('Enter User A username (1min streaming /wo bid-ask) - ')
    pass_A = input('Enter password for User A - ')
    user_B = input('Enter User B username (tick streaming /w bid-ask) - ')
    pass_B = input('Enter password for User B - ')
    live_port = None
    hist_port = None
    symbols = None
    try:
        for arg in sys.argv[2:]:
            if arg[:2] == "--":
                arg_split = arg[2:].split('=')
                arg_key = arg_split[0]
                arg_value = arg_split[1]
                if arg_key.lower() == 'liveport':
                    live_port = int(arg_value)
                if arg_key.lower() == 'histport':
                    hist_port = int(arg_value)
                if arg_key.lower() == 'symbols':
                    symbols = arg_value.split()
                    symbols = [symbol.strip() for symbol in symbols]
    except IndexError:
        pass
    internal_tests.run_all_tests(user_A, pass_A, user_B, pass_B, live_port, hist_port, symbols)



from .TD import TD

from copy import deepcopy
from time import sleep

import pandas as pd
from datetime import datetime
from dateutil.relativedelta import relativedelta

from colorama import Style, Fore


def get_connection(username, password, live_port_ip, historical_port_ip):
    td_obj = TD(username, password, live_port=live_port_ip, historical_port=historical_port_ip)
    return td_obj


def run_historical_tests(td_obj, symbols):
    symbol = symbols[0]
    test_time = datetime.today()
    # Testing malformed historical contracts
    # hist_data_0 = td_obj.get_historic_data(f'BANKNIFTY99ZYXFUT')

    # Testing bar historical data
    hist_data_1 = td_obj.get_historic_data(f'{symbol}')
    hist_data_2 = td_obj.get_historic_data(f'{symbol}', duration='3 D')
    hist_data_3 = td_obj.get_historic_data(f'{symbol}', duration='3 D', bar_size='15 mins')
    hist_data_4 = td_obj.get_historic_data(f'{symbol}', bar_size='30 mins')
    hist_data_5 = td_obj.get_historic_data(f'{symbol}', bar_size='30 mins', start_time=test_time - relativedelta(days=3))
    hist_data_6 = td_obj.get_historic_data(f'{symbol}', bar_size='EOD', duration='1 M')
    # Testing tick historical data
    tick_hist_data_1 = td_obj.get_historic_data(f'{symbol}', bar_size='tick')
    tick_hist_data_2 = td_obj.get_historic_data(f'{symbol}', bar_size='tick', duration='3 D')
    tick_hist_data_3 = td_obj.get_historic_data(f'{symbol}', bar_size='tick', start_time=test_time - relativedelta(days=3))

    print(f'{Style.BRIGHT}{Fore.BLUE}------------- HIST BAR DATA TEST RESULTS -------------{Style.RESET_ALL}')
    print()
    print(f"{Style.BRIGHT}{Fore.BLUE}HISTDATA 1...\n"
          f"\tCommand used -> hist_data_1 = td_app.get_historic_data('{symbol}')\n"
          f"\tLENGTH OF RESULT = {len(hist_data_1)}{Style.RESET_ALL}")
    for hist_point in hist_data_1[-20:]:
        print(hist_point)
    print()
    print(f"{Style.BRIGHT}{Fore.BLUE}HISTDATA 2...\n"
          f"\tCommand used -> hist_data_2 = td_app.get_historic_data('{symbol}', duration='3 D')\n"
          f"\tLENGTH OF RESULT = {len(hist_data_2)}{Style.RESET_ALL}")
    for hist_point in hist_data_2[-20:]:
        print(hist_point)
    print()
    print(f"{Style.BRIGHT}{Fore.BLUE}HISTDATA 3...\n"
          f"\tCommand used -> hist_data_3 = td_app.get_historic_data('{symbol}', duration='3 D', bar_size='15 mins')\n"
          f"\tLENGTH OF RESULT = {len(hist_data_3)}{Style.RESET_ALL}")
    for hist_point in hist_data_3[-20:]:
        print(hist_point)
    print()
    print(f"{Style.BRIGHT}{Fore.BLUE}HISTDATA 4...\n"
          f"\tCommand used -> hist_data_4 = td_app.get_historic_data('{symbol}', bar_size='30 mins')\n"
          f"\tLENGTH OF RESULT = {len(hist_data_4)}{Style.RESET_ALL}")
    for hist_point in hist_data_4[-20:]:
        print(hist_point)
    print()
    print(f"{Style.BRIGHT}{Fore.BLUE}HISTDATA 5...\n"
          f"\tCommand used -> hist_data_5 = td_app.get_historic_data('{symbol}', bar_size='30 mins', start_time=datetime({(test_time - relativedelta(days=3)).strftime('%Y, %m, %d, %H, %M, %S').replace(' 0', ' ')}))\n"
          f"\tLENGTH OF RESULT = {len(hist_data_5)}{Style.RESET_ALL}")
    for hist_point in hist_data_5[-20:]:
        print(hist_point)
    print()
    print(f"{Style.BRIGHT}{Fore.BLUE}HISTDATA 6...\n"
          f"\tCommand used -> hist_data_6 = td_obj.get_historic_data(f'{symbol}', bar_size='EOD', duration='1 M'))\n"
          f"\tLENGTH OF RESULT = {len(hist_data_6)}{Style.RESET_ALL}")
    for hist_point in hist_data_6[-20:]:
        print(hist_point)
    print()
    print()
    print(f'{Style.BRIGHT}{Fore.BLUE}------------- HIST TICK DATA TEST RESULTS -------------{Style.RESET_ALL}')
    print()
    print(f"{Style.BRIGHT}{Fore.BLUE}TICKDATA 1...\n"
          f"\tCommand used -> tick_data_1 = td_app.get_historic_data('{symbol}', bar_size='tick')\n"
          f"\tLENGTH OF RESULT = {len(tick_hist_data_1)}{Style.RESET_ALL}")
    for hist_point in tick_hist_data_1[-20:]:
        print(hist_point)
    print()
    print(f"{Style.BRIGHT}{Fore.BLUE}TICKDATA 2...\n"
          f"\tCommand used -> tick_data_2 = td_app.get_historic_data('{symbol}', bar_size='tick', duration='3 D')\n"
          f"\tLENGTH OF RESULT = {len(tick_hist_data_2)}{Style.RESET_ALL}")
    for hist_point in tick_hist_data_2[-20:]:
        print(hist_point)
    print()
    print(f"{Style.BRIGHT}{Fore.BLUE}TICKDATA 3...\n"
          f"\tCommand used -> tick_data_3 = td_app.get_historic_data('{symbol}', bar_size='tick', start_time=datetime({(test_time - relativedelta(days=3)).strftime('%Y, %m, %d, %H, %M, %S').replace(' 0', ' ')}))\n"
          f"\tLENGTH OF RESULT = {len(tick_hist_data_3)}{Style.RESET_ALL}")
    for hist_point in tick_hist_data_3[-20:]:
        print(hist_point)

    # Testing conversion to pandas dataframe
    print(f'{Style.BRIGHT}{Fore.BLUE}Converting HISTDATA 1 to a Pandas DataFrame{Style.RESET_ALL}')
    print(f'Command used -> df = pd.DataFrame(hist_data_1)')
    df = pd.DataFrame(hist_data_1)
    print(df)
    print(f'{Style.BRIGHT}{Fore.BLUE}Converting TICKDATA 1 to a Pandas DataFrame{Style.RESET_ALL}')
    print(f'Command used -> df = pd.DataFrame(tick_hist_data_1)')
    df = pd.DataFrame(tick_hist_data_1)
    print(df)


def run_live_tests(td_obj, symbols):
    # Testing Live data
    # td_app.start_live_data(f'BANKNIFTY{sys.argv[3]}FUTbk')
    print(f'{Style.BRIGHT}{Fore.BLUE}Checking LIVE data streaming...{Style.RESET_ALL}')
    req_ids = td_obj.start_live_data(symbols)
    sleep(3)
    print(f"{Style.BRIGHT}{Fore.BLUE}Here's the touchline data...{Style.RESET_ALL}")
    for req_id in req_ids:
        print(td_obj.touchline_data[req_id])
    print()
    print(f"{Style.BRIGHT}{Fore.BLUE}Here's the LIVE stream... Use Ctrl+C to stop and exit...{Style.RESET_ALL}")
    sleep(2)  # This is here just to give the user time to read the instructions...
    live_data_objs = {}
    for req_id in req_ids:
        live_data_objs[req_id] = deepcopy(td_obj.live_data[req_id])
    while True:
        try:
            for req_id in req_ids:
                if live_data_objs[req_id] != td_obj.live_data[req_id]:
                    print(td_obj.live_data[req_id])
                    live_data_objs[req_id] = deepcopy(td_obj.live_data[req_id])
        except KeyboardInterrupt:
            td_obj.stop_live_data(symbols)
            td_obj.disconnect()
            exit()



from io import StringIO
from copy import deepcopy
from datetime import datetime as dt
from datetime import time as t

import pandas as pd
import requests
import re
import numpy as np
import time



class OptionChain:

    chain_url ='https://api.truedata.in/getOptionChain?'
    close_time = t(15, 40)

    def __init__(self,
                 TD_OBJ ,
                 symbol ,
                 expiry ,
                 chain_length ,
                 future_price,
                 bid_ask ,
                 market_open_post_hours ):
        self.TD_OBJ = TD_OBJ
        self.login_id = TD_OBJ.login_id
        self.password = TD_OBJ.password
        self.symbol = symbol
        self.expiry = expiry
        self.chain_length = chain_length
        self.future_price = future_price
        self.strike_step = self.get_strike_step()
        self.option_symbols = self.get_option_symbols()
        self.chain_dataframe =  self.init_dataframe()
        self.chain_status = False
        self.bid_ask = bid_ask
        self.market_open_post_hours = market_open_post_hours

    def get_strike_step(self ):
        expiry = self.expiry.strftime('%Y%m%d')
        chain_link = f'{OptionChain.chain_url}user={self.login_id}&password={self.password}&symbol={self.symbol}&expiry={expiry}&csv=true'
        chain = requests.get(chain_link).text
        df = pd.read_csv(StringIO(chain) , header = None)
        df[1] = df[1].apply(lambda x : re.findall('\D+' , x)[0])
        df = df[df[1] == self.symbol]
        strikes = np.sort(df[6].unique())
        step = strikes[1]-strikes[0];
        for i in range(2 , len(strikes)):
            step = min(step, strikes[i]-strikes[i-1]);
        return step

    def get_option_symbols(self ):
        expiry = self.expiry.strftime('%y%m%d')
        future_atm = round(self.future_price / self.strike_step ) * self.strike_step
        start_strike = future_atm - self.strike_step * int(self.chain_length / 2 )
        end_strike = future_atm + self.strike_step * int( self.chain_length / 2)
        req_strikes = np.arange( start_strike , end_strike  , self.strike_step )
        if isinstance( self.strike_step , float ) :
            req_strikes = list(map(lambda x: int(x) if x.is_integer() else x , req_strikes))
        symbols = list(map(lambda x: f'{self.symbol}{expiry}{x}CE', req_strikes))
        symbols.extend(list(map(lambda x: f'{self.symbol}{expiry}{x}PE', req_strikes )))
        return symbols

    def init_dataframe(self):
        df = pd.DataFrame(columns=['symbols' , 'strike' , 'type' , 'ltp' , 'ltt' , 'ltq' , 'volume' , 'price_change' , 'price_change_perc', 'oi' , 
                                    'prev_oi' , 'oi_change' , 'oi_change_perc' , 'bid' , 'bid_qty' , 'ask' , 'ask_qty'])
        df.symbols = self.option_symbols
        df.strike = df.symbols.apply(lambda x: str(re.search(r"\d+(\.\d+)?", x ).group(0) )[6:])
        df.type = df.symbols.apply(lambda x: re.findall('\D+' , x )[-1] )
        df.set_index('symbols' , inplace=True)
        df.sort_values(by=['strike' , 'type'] , inplace=True)
        return df

    def initial_update_from_touchline(self , td_obj , symb_ids):
        prev_tick_obj = {}
        time.sleep(1)
        for symb_id in symb_ids:
            prev_tick_obj[symb_id] = deepcopy(td_obj.live_data[symb_id])
            try:
                oi_change_perc = round((td_obj.touchline_data[symb_id].oi - td_obj.touchline_data[symb_id].prev_oi)*100 / (td_obj.touchline_data[symb_id].prev_oi) , 2 )
                price_change_perc = round((td_obj.touchline_data[symb_id].ltp - td_obj.touchline_data[symb_id].prev_close)*100 / (td_obj.touchline_data[symb_id].prev_oi) , 4 )
            except (ZeroDivisionError , TypeError):
                oi_change_perc = 0
                price_change_perc = 0
            try:
                self.chain_dataframe.loc[td_obj.touchline_data[symb_id].symbol , ['ltp' , 'ltt'  , 'ltq' , 'volume' ,'price_change' , 'price_change_perc' , 'oi' , 
                                                        'prev_oi', 'oi_change' , 'oi_change_perc','bid' , 'bid_qty' , 'ask' , 'ask_qty' ]] = [
                    td_obj.touchline_data[symb_id].ltp, td_obj.touchline_data[symb_id].timestamp , td_obj.touchline_data[symb_id].ltq , 
                    td_obj.touchline_data[symb_id].ttq ,td_obj.touchline_data[symb_id].ltp - td_obj.touchline_data[symb_id].prev_close , 
                    price_change_perc , td_obj.touchline_data[symb_id].oi, td_obj.touchline_data[symb_id].prev_oi , 
                    td_obj.touchline_data[symb_id].oi - td_obj.touchline_data[symb_id].prev_oi ,oi_change_perc , 
                    td_obj.touchline_data[symb_id].best_bid_price, td_obj.touchline_data[symb_id].best_bid_qty,
                    td_obj.touchline_data[symb_id].best_ask_price, td_obj.touchline_data[symb_id].best_ask_qty 
                ]
            except TypeError:
                self.chain_dataframe.loc[td_obj.touchline_data[symb_id].symbol , ['ltp' , 'ltt' ,'ltq' , 'volume' , 'price_change' , 'price_change_perc',  'oi' , 'prev_oi', 'oi_change' ,
                                                            'oi_change_perc','bid' , 'bid_qty' , 'ask' , 'ask_qty' ]] =[ np.NaN for i in range(14)]
        return prev_tick_obj

    def update_chain(self ,td_obj , symb_ids ):
        self.chain_status = True
        prev_tick_obj = self.initial_update_from_touchline(td_obj, symb_ids)
        while self.chain_status:
            for symb_id in symb_ids:
                if not td_obj.live_data[symb_id] == prev_tick_obj[symb_id]:
                    try:
                        bid, bid_qnty = td_obj.live_data[symb_id].best_bid_price , td_obj.live_data[symb_id].best_bid_qty
                        ask , ask_qnty = td_obj.live_data[symb_id].best_ask_price, td_obj.live_data[symb_id].best_ask_qty
                    except Exception as e:
                        bid, bid_qnty, ask, ask_qnty = 0, 0, 0, 0
                    self.chain_dataframe.loc[td_obj.live_data[symb_id].symbol , ['ltt' , 'volume' , 'price_change','price_change_perc', 'oi' ,'prev_oi' , 'oi_change' ,
                                                                                'oi_change_perc' ,'bid' , 'bid_qty' , 'ask' , 'ask_qty' ]] = [
                        td_obj.live_data[symb_id].timestamp , td_obj.live_data[symb_id].ttq ,td_obj.live_data[symb_id].change, 
                        round(td_obj.live_data[symb_id].change_perc , 2) ,td_obj.live_data[symb_id].oi, td_obj.live_data[symb_id].prev_day_oi ,
                        td_obj.live_data[symb_id].oi_change ,round(td_obj.live_data[symb_id].to_dict()['oi_change_perc'] ,2 ) ,bid, bid_qnty, ask, ask_qnty 
                    ]
                    if 'ltp' in td_obj.live_data[symb_id].__dict__.keys():
                        self.chain_dataframe.loc[td_obj.live_data[symb_id].symbol , ['ltp'] ] = td_obj.live_data[symb_id].ltp    
                        self.chain_dataframe.loc[td_obj.live_data[symb_id].symbol , ['ltq'] ] = td_obj.live_data[symb_id].ltq if td_obj.live_data[symb_id].ltq is not None else 0
                    else:
                        self.chain_dataframe.loc[td_obj.live_data[symb_id].symbol , ['ltp'] ] = td_obj.live_data[symb_id].close
                        self.chain_dataframe.loc[td_obj.live_data[symb_id].symbol , ['ltq'] ] = td_obj.live_data[symb_id].volume if td_obj.live_data[symb_id].volume is not None else 0
                    prev_tick_obj[symb_id] = deepcopy(td_obj.live_data[symb_id])
            time.sleep(0.005)
            if not(self.market_open_post_hours) and dt.now().time() > OptionChain.close_time :
                break
        # print('exited chain thread')

    def get_option_chain(self):
        df = self.chain_dataframe.copy()
        df.dropna(inplace=True)
        df = df.astype({'ltq' : int , 'volume': int })
        if not self.bid_ask:
            df = df.drop(['bid' , 'bid_qty' , 'ask' , 'ask_qty'] , axis= 1  )
        return df

    def stop_option_chain(self):
        self.chain_status = False
        self.TD_OBJ.stop_live_data(self.option_symbols)




# Python Version Defaults
MIN_PYTHON_MAJOR = 3
MIN_PYTHON_MINOR = 7

# Testing Defaults
DEFAULT_LIVE_PORT = 8082
DEFAULT_HIST_PORT = 8092
DEFAULT_TEST_SYMBOL_LIST = ['BANKNIFTY-I', 'NIFTY-I', 'CRUDEOIL-I']



from .support import historical_decorator, access_token_decorator
from io import StringIO
from datetime import datetime, timedelta
from logging import Logger
from threading import RLock
from colorama import Style, Fore
from typing import List, Dict

import requests
import pandas as pd
import json
import lz4.block
import struct
import base64

class HistoricalREST:
    def __init__(self, login_id: str, password: str, url: str, broker_token: str, logger: Logger):  # NO PORT, broker token needed from now on
        self.login_id = login_id
        self.password = password
        self.url = url
        self.broker_token = broker_token
        self.logger = logger
        self.thread_lock = RLock()
        self.access_token = None
        self.bhavcopy_last_completed = None
        self.access_token_expiry_time = None
        try:
            self.hist_login()
        except Exception as e:
            self.logger.error(f"Failed to connect REST historical API -> {type(e)} = {e}")

    def hist_login(self):
        url_auth = "https://auth.truedata.in/token"
        payload = f"username={self.login_id}&password={self.password}&grant_type=password"
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded'
        }
        welcome_msg = requests.request("POST", url_auth, headers=headers, data=payload).text
        welcome_msg = json.loads(welcome_msg)
        try:
            if welcome_msg['access_token']:
                self.access_token = welcome_msg['access_token']
                self.access_token_expiry_time = datetime.now() + timedelta(seconds=welcome_msg['expires_in'] - 15)  # 15 seconds is random buffer time
                self.logger.warning(f"{Style.NORMAL}{Fore.BLUE}Connected successfully to TrueData Historical Data Service... {Style.RESET_ALL}")
        except Exception as e:
            self.logger.error(f"Failed to connect -> {welcome_msg['error_description']}{type(e)} = {e}")
            self.access_token = None

    # noinspection PyUnusedLocal
    @access_token_decorator
    @historical_decorator
    def get_n_historic_bars(self, contract, end_time, no_of_bars, bar_size, options=None, bidask=False):
        source_bar_type_string = 'getlastnbars'
        if 'data_type' not in options.keys():
            options['data_type'] = 'csv'
        try:
            encoded_payload = {
                'symbol': contract,
                'interval': bar_size,
                'response': options['data_type'],
                'bidask': 0 ,
                'comp': 'true'
            }
            unencoded_payload = {}
            unencoded_payload = "&".join(unencoded_payload)
            if bar_size == 'tick':
                encoded_payload['nticks'] = no_of_bars
                source_bar_type_string = 'getlastnticks'
                if bidask:
                    encoded_payload['bidask'] = 1
            else:  # Not ticks
                encoded_payload['nbars'] = no_of_bars
            headers = {
                'Authorization': f'Bearer {self.access_token}'
            }

            with self.thread_lock:
                url = f"{self.url}/{source_bar_type_string}?{unencoded_payload}"
                response = requests.get(url, headers=headers, params=encoded_payload)
                if response.status_code == 429:
                    error = base64.b64decode(response.content).decode('utf-8')
                    raise TooManyRequestsError(error)
                hist_data = HistoricalREST.decompress_data(response.content)
        except Exception as e:
            self.logger.error(f"{type(e)} -> {e}")
            return None
            # json_response = {"success": False, "message": f"({type(e)}) {addn_err_str} -> {e}"}
        hist_data = self.parse_data(hist_data, options)
        return options['processor_to_call'][options['data_type']](hist_data, time_format=options['time_format'])

    @access_token_decorator
    @historical_decorator
    def get_historic_data(self, contract, end_time, start_time, bar_size, options=None, bidask=False):
        source_bar_type_string = 'getbars'
        if 'data_type' not in options.keys():
            options['data_type'] = 'csv'
        try:
            encoded_payload = {
                'symbol': contract,
                'interval': bar_size,
                'response': options['data_type'],
                'comp': 'true'
            }
            unencoded_payload = {
                f"from={start_time}",
                f"to={end_time}"
            }
            unencoded_payload = "&".join(unencoded_payload)
            if bar_size == 'tick':
                encoded_payload['bidask'] = 0
                source_bar_type_string = 'getticks'
                if bidask:
                    encoded_payload['bidask'] = 1
            headers = {
                'Authorization': f'Bearer {self.access_token}'
            }
            with self.thread_lock:
                url = f"{self.url}/{source_bar_type_string}?{unencoded_payload}"
                response = requests.get(url, headers=headers, params=encoded_payload)
                if response.status_code == 429:
                    error = base64.b64decode(response.content).decode('utf-8')
                    raise TooManyRequestsError(error)
                hist_data = HistoricalREST.decompress_data(response.content)
        except Exception as e:
            self.logger.error(f"{type(e)} -> {e}")
            return None
            # json_response = {"success": False, "message": f"({type(e)}) {addn_err_str} -> {e}"}
        hist_data = self.parse_data(hist_data, options)
        return options['processor_to_call'][options['data_type']](hist_data, time_format=options['time_format'])


    def get_gainers_losers(self , segment, topn , gainers , df_style ):
        source_string = "gettopngainers" if gainers else "gettopnlosers"
        headers = {'Authorization': f'Bearer {self.access_token}'}
        payload = {'segment': segment , 'topn': topn , 'response':'csv' }
        url = f"{self.url}/{source_string}?"
        try:
            with self.thread_lock:
                response = requests.get(url, headers=headers , params=payload )
                out = response.text
            df = pd.read_csv(StringIO(out))
            df = df if df_style else list(df.T.to_dict().values())
            df.set_index(["symbol"] , inplace=True) if df_style else 0
        except Exception as e:
            self.logger.error(f"{type(e)} -> No match found for this segment {segment}")
            df = []
        return df


    def parse_data(self, hist_data, options):
        if options['data_type'] == 'csv':
            if hist_data.startswith('No data exists for') or hist_data.startswith('Symbol does not exist'):
                self.logger.error(f'{hist_data}')
                return ''  # Note to devs: Leave this as str to avoid "split" error
            else:
                return hist_data
        else:  # JSON format
            try:
                hist_data = json.loads(hist_data)
            except json.decoder.JSONDecodeError:
                self.logger.error(f'Caught a JSONDecodeError for this request')
                return []
            if hist_data['status'] != 'Success':
                self.logger.error(f"{hist_data['status']}")
                return []
            else:
                return hist_data['Records']

    def bhavcopy_status(self, segment: str):
        url = f'https://history.truedata.in/getbhavcopystatus?segment={segment}&response=csv'
        payload = {}
        headers = {
            'Authorization': f'Bearer {self.access_token}'
        }
        response = requests.request("GET", url, headers=headers, data=payload)
        status = response.text.strip().split('\r\n')[-1].split(',')
        assert segment == status[0]
        self.bhavcopy_last_completed = datetime.strptime(status[1], "%Y-%m-%dT%H:%M:%S")

    @access_token_decorator
    def bhavcopy(self, segment: str, date: datetime, return_completed: bool) -> List[Dict]:
        try:
            self.bhavcopy_status(segment)
            if return_completed:
                if date > self.bhavcopy_last_completed:
                    self.logger.error(f"{Style.BRIGHT}{Fore.RED}No complete bhavcopy found for requested date."
                                      f" Last available for {self.bhavcopy_last_completed.strftime('%Y-%m-%d %H:%M:%S')}.{Style.RESET_ALL}")
                    return []
            url_bhavcopy = f"https://history.truedata.in/getbhavcopy?segment={segment}&date={date.strftime('%Y-%m-%d')}&response=csv"
            payload = { 'comp' : 'true' }
            headers = {
                'Authorization': f'Bearer {self.access_token}'
            }
            response = requests.request( "GET", url_bhavcopy, headers=headers, params=payload )
            response = HistoricalREST.decompress_data( response.content )
            bhavcopy_data = response.strip().split('\r\n')
            data_list = []
            # headers = hist_data[0]
            bhavcopy_data = bhavcopy_data[1:]
            for j in bhavcopy_data:
                j = j.split(',')
                data_list.append({'symbol_id': int(j[0]),
                                  'symbol': str(j[1]),
                                  'date': datetime.strptime(str(j[2]), '%Y-%m-%d'),
                                  'o': float(j[3]),
                                  'h': float(j[4]),
                                  'l': float(j[5]),
                                  'c': float(j[6]),
                                  'v': int(j[7]),
                                  'oi': int(j[8])})
            return data_list
        except Exception as e:
            self.logger.error(f"{type(e)} -> {e}")

    @staticmethod
    def hist_json_tick_data_to_dict_list(hist_data, time_format):
        data_list = []
        for j in hist_data:
            try:
                data_list.append({'time': datetime.strptime(j[0], time_format),
                                  'ltp': float(j[1]),
                                  'volume': int(j[2]),
                                  'oi': int(j[3] if len(j[3]) >= 1 else 0),
                                  'bid': float(j[4]),
                                  'bid_qty': int(j[5]),
                                  'ask': float(j[6]),
                                  'ask_qty': int(j[7])})
            except IndexError:  # No bid-ask data
                data_list.append({'time': datetime.strptime(j[0], time_format),
                                  'ltp': float(j[1]),
                                  'volume': int(j[2]),
                                  'oi': int(j[3] if len(j[3]) >= 1 else 0)})
                continue
        return data_list

    @staticmethod
    def hist_json_bar_data_to_dict_list(hist_data, time_format):
        data_list = []
        for j in hist_data:
            data_list.append({'time': datetime.strptime(j[0], time_format),
                              'o': float(j[1]),
                              'h': float(j[2]),
                              'l': float(j[3]),
                              'c': float(j[4]),
                              'v': int(j[5]),
                              'oi': int(j[6])})
        return data_list

    @staticmethod
    def hist_csv_tick_data_to_dict_list(hist_data, time_format):
        hist_data = hist_data.split()
        data_list = []
        # headers = hist_data[0]
        hist_data = hist_data[1:]
        count = 0
        for j in hist_data:
            j = j.split(',')
            try:
                data_list.append({'time': datetime.strptime(j[0], time_format),
                                  'ltp': float(j[1]),
                                  'volume': int(j[2]),
                                  'oi': int(j[3] if len(j[3]) >= 1 else 0),
                                  'bid': float(j[4]),
                                  'bid_qty': int(j[5]),
                                  'ask': float(j[6]),
                                  'ask_qty': int(j[7])})
            except IndexError:  # No bid-ask data
                data_list.append({'time': datetime.strptime(j[0], time_format),
                                  'ltp': float(j[1]),
                                  'volume': int(j[2]),
                                  'oi': int(j[3] if len(j[3]) >= 1 else 0)})
                continue
            count = count + 1
        return data_list

    @staticmethod
    def hist_csv_bar_data_to_dict_list(hist_data, time_format):
        hist_data = hist_data.split()
        data_list = []
        # headers = hist_data[0]
        hist_data = hist_data[1:]
        for j in hist_data:
            j = j.split(',')
            # time_format = '%Y-%m-%dT%H:%M:%S'
            data_list.append({'time': datetime.strptime(j[0], time_format),
                              'o': float(j[1]),
                              'h': float(j[2]),
                              'l': float(j[3]),
                              'c': float(j[4]),
                              'v': int(j[5]),
                              'oi': int(j[6])})
        return data_list

    @staticmethod
    def decompress_data(data):
        uncom_length = struct.unpack('<I', data[:4])[0]
        com_length = struct.unpack('<I', data[4:8])[0]
        dc = lz4.block.decompress( data[8:], uncom_length ) if com_length != uncom_length else data[8:]
        return dc.decode()

class TooManyRequestsError(Exception):
    def __init__(self, msg):
        self.msg = msg
    def __str__(self ):
        return self.msg




from websocket import WebSocketApp
from threading import Thread
from datetime import datetime
from dateutil.relativedelta import relativedelta
import time
import json
from copy import deepcopy
from collections import defaultdict

from colorama import Style, Fore


class LiveClient(WebSocketApp):

    def __init__(self, parent_app, url, *args):
        WebSocketApp.__init__(self, url, on_open=self.ws_open, on_error=self.ws_error, on_message=self.on_msg_func, on_close=self.ws_close, *args)
        self.segments = []
        self.max_symbols = 0
        self.remaining_symbols = 0
        self.valid_until = ''
        self.contract_mapping = {}
        self.subscription_type = ''
        self.confirm_heartbeats = 1
        self.store_last_n_heartbeats = self.confirm_heartbeats + 7
        self.heartbeat_interval = 5
        self.heartbeat_buffer = 0.5
        time_of_creation = datetime.now()
        self.last_n_heartbeat = [time_of_creation - relativedelta(seconds=i * self.heartbeat_interval) for i in range(-self.store_last_n_heartbeats, 0)]
        self.parent_app = parent_app
        self.logger = self.parent_app.logger
        self.disconnect_flag = False
        self.heartbeat_check_thread = Thread(target=self.check_heartbeat, daemon=True)
        self.trade_callback = None
        self.bar_callback = None
        self.bidask_callback = None
        # if self.parent_app.live_port == 8086 or self.parent_app.live_port == 8084:
        #     # self.heartbeat_check_thread.start()
        #     pass

    def check_connection(self):
        base_heartbeat = self.last_n_heartbeat[-self.confirm_heartbeats]
        check_time = datetime.now()
        time_diff = check_time - base_heartbeat
        is_connected = time_diff.total_seconds() > ((self.heartbeat_interval + self.heartbeat_buffer) * self.confirm_heartbeats)  # 3 > 5 + 0.5
        return is_connected

    def check_heartbeat(self):
        while True:
            time.sleep(self.heartbeat_interval)
            if self.disconnect_flag:
                self.logger.info(f"{Fore.WHITE}Removing hand from the pulse...{Style.RESET_ALL}")
                break
            if self.check_connection():
                self.logger.debug(f"{Style.BRIGHT}{Fore.RED}Failed Heartbeat @ {datetime.now()} because of last at {self.last_n_heartbeat[-self.confirm_heartbeats]}{Style.RESET_ALL}")
                self.logger.info(f"{Style.BRIGHT}{Fore.RED}Attempting reconnect @ {Fore.CYAN}{datetime.now()}{Style.RESET_ALL}")
                restart_successful = self.reconnect()
                if restart_successful:
                    self.logger.info(f"{Style.BRIGHT}{Fore.GREEN}Successful restart @ {Fore.CYAN}{datetime.now()}{Style.RESET_ALL}")
                    time.sleep((self.heartbeat_interval + self.heartbeat_buffer))
                    recover_start, recover_end = self.get_largest_diff(self.last_n_heartbeat)
                    # self.logger.debug(f"\t\t\t{len(self.last_n_heartbeat)} - {self.last_n_heartbeat}")
                    self.recover_from_time_missed(recover_start, recover_end)
                else:
                    self.logger.info(f"{Style.BRIGHT}{Fore.RED}Failed restart @ {Fore.CYAN}{datetime.now()}{Style.RESET_ALL}")

    @staticmethod
    def get_largest_diff(time_series):
        big_li = deepcopy(time_series.pop(0))
        small_li = deepcopy(time_series.pop())
        diffs = [i[0]-i[1] for i in zip(big_li, small_li)]
        start_gap_index = max(range(len(diffs)), key=lambda i: diffs[i])
        return time_series[start_gap_index], time_series[start_gap_index+1]

    def recover_from_time_missed(self, start_time, end_time):
        self.logger.info(f"{Style.BRIGHT}{Fore.YELLOW}Initiating recovery from {Fore.GREEN}{start_time}{Fore.YELLOW} till {Fore.GREEN}{end_time}{Fore.YELLOW} "
                         f"which are last green heartbeats from server...{Style.RESET_ALL}")

    def reconnect(self):
        self.close()
        t = Thread(target=self.run_forever, daemon=True)
        t.start()
        time.sleep(1)
        is_td_connected = False
        while not is_td_connected:
            time.sleep(self.heartbeat_interval + self.heartbeat_buffer)
            is_td_connected = self.check_connection()
        return is_td_connected

    def on_msg_func(self, *args):
        message = args[-1]
        msg = json.loads(message)
        if 'message' in msg.keys():
            self.handle_message_data(msg)
        if 'trade' in msg.keys():
            trade = msg['trade']
            self.handle_trade_data(trade)
        elif 'bidask' in msg.keys():
            bidask = msg['bidask']
            self.handle_bid_ask_data(bidask)
        elif any(['min' in key for key in msg.keys()]):
            bar_key = next(key for key in msg.keys() if 'min' in key)
            bar_data = msg[bar_key]
            self.handle_bar_data(bar_data)

    def handle_message_data(self, msg):
        if msg['success']:
            if msg['message'] == 'HeartBeat':
                self.handle_heartbeat(msg['timestamp'])
            elif msg['message'] == 'TrueData Real Time Data Service':  # Connection success message
                # self.logger.info(f"{Style.BRIGHT}{Fore.WHITE}You have subscribed for {msg['maxsymbols']} "
                #                  f"symbols across {msg['segments']} until {msg['validity']} with type of stream "
                #                  f"as {msg['subscription']}...{Style.RESET_ALL}")
                self.logger.warning(f"{Style.NORMAL}{Fore.BLUE}Connected successfully to {msg['message']}... {Style.RESET_ALL}")
                self.subscription_type = msg['subscription']
            elif msg['message'] in ['symbols added', 'touchline']:
                self.add_contract_details(msg['symbollist'])
            elif msg['message'] == 'symbols removed':
                self.remove_symbols(msg['symbollist'])
        else:
            self.logger.error(f"{Style.BRIGHT}{Fore.RED}The request encountered an error - {msg['message']}{Style.RESET_ALL}")

    def handle_heartbeat(self, server_timestamp):
        self.logger.debug(f'Server heartbeat received at {server_timestamp}')
        timestamp = datetime.strptime(server_timestamp, "%Y-%m-%dT%H:%M:%S.%f")  # old format + ((26 - len(server_timestamp)) * "0")
        self.last_n_heartbeat = self.last_n_heartbeat[1:]
        self.last_n_heartbeat.append(timestamp)

    def remove_symbols(self, contracts):
        for contract_info in contracts:
            contract = contract_info.split(':')[0]
            for req_id in self.parent_app.symbol_mkt_id_map[contract.upper()]:
                del self.parent_app.live_data[req_id]
                del self.parent_app.touchline_data[req_id]
                if '+' in self.subscription_type:
                    del self.parent_app.min_live_data[req_id]
            del self.parent_app.symbol_mkt_id_map[contract.upper()]

    def add_contract_details(self, contracts_list):
        for contract_details in contracts_list:
            if contract_details is not None:
                self.contract_mapping[int(contract_details[1])] = symbol = contract_details[0]
                for ticker_id in self.parent_app.symbol_mkt_id_map[symbol]:
                    self.parent_app.touchline_data[ticker_id].symbol_id = int(contract_details[1])
                    self.parent_app.touchline_data[ticker_id].timestamp = datetime.strptime(contract_details[2], '%Y-%m-%dT%H:%M:%S')
                    self.parent_app.touchline_data[ticker_id].symbol = symbol
                    self.parent_app.touchline_data[ticker_id].ltp = float(contract_details[3])
                    self.parent_app.touchline_data[ticker_id].ltq = float(contract_details[4])
                    self.parent_app.touchline_data[ticker_id].atp = float(contract_details[5])
                    self.parent_app.touchline_data[ticker_id].ttq = int(contract_details[6])
                    self.parent_app.touchline_data[ticker_id].open = float(contract_details[7])
                    self.parent_app.touchline_data[ticker_id].high = float(contract_details[8])
                    self.parent_app.touchline_data[ticker_id].low = float(contract_details[9])
                    self.parent_app.touchline_data[ticker_id].prev_close = float(contract_details[10])
                    self.parent_app.touchline_data[ticker_id].oi = int(contract_details[11])
                    self.parent_app.touchline_data[ticker_id].prev_oi = int(contract_details[12])
                    self.parent_app.touchline_data[ticker_id].turnover = float(contract_details[13])
                    self.parent_app.touchline_data[ticker_id].best_bid_price = float(contract_details[14])
                    self.parent_app.touchline_data[ticker_id].best_bid_qty = float(contract_details[15])
                    self.parent_app.touchline_data[ticker_id].best_ask_price = float(contract_details[16])
                    self.parent_app.touchline_data[ticker_id].best_ask_qty = float(contract_details[17])
                    self.parent_app.live_data[ticker_id].populate_using_touchline(self.parent_app.live_data[ticker_id], self.parent_app.touchline_data[ticker_id])
                    if '+' in self.subscription_type:
                        self.parent_app.min_live_data[ticker_id].populate_using_touchline(self.parent_app.min_live_data[ticker_id], self.parent_app.touchline_data[ticker_id])
                self.logger.info(f"Updated touchline data for {symbol} and related live data objects...")
            else:
                self.logger.debug(f'{Style.BRIGHT}{Fore.YELLOW}Probable repeated symbol...{Style.RESET_ALL}')

    def handle_trade_data(self, trade_tick):
        try:
            symbol = self.contract_mapping[int(trade_tick[0])]
            for ticker_id in self.parent_app.symbol_mkt_id_map[symbol]:
                self.parent_app.live_data[ticker_id].symbol_id = int(trade_tick[0])
                self.parent_app.live_data[ticker_id].timestamp = datetime.strptime(trade_tick[1], '%Y-%m-%dT%H:%M:%S')
                self.parent_app.live_data[ticker_id].symbol = symbol
                self.parent_app.live_data[ticker_id].ltp = self.parent_app.touchline_data[ticker_id].ltp = ltp = float(trade_tick[2])
                self.parent_app.live_data[ticker_id].ltq = float(trade_tick[3])
                self.parent_app.live_data[ticker_id].atp = float(trade_tick[4])
                self.parent_app.live_data[ticker_id].ttq = float(trade_tick[5])
                self.parent_app.live_data[ticker_id].day_open = self.parent_app.touchline_data[ticker_id].open = float(trade_tick[6])
                self.parent_app.live_data[ticker_id].day_high = float(trade_tick[7])
                self.parent_app.live_data[ticker_id].day_low = float(trade_tick[8])
                self.parent_app.live_data[ticker_id].prev_day_close = float(trade_tick[9])
                self.parent_app.live_data[ticker_id].oi = int(trade_tick[10])
                self.parent_app.live_data[ticker_id].prev_day_oi = int(trade_tick[11])
                self.parent_app.live_data[ticker_id].turnover = float(trade_tick[12])
                self.parent_app.live_data[ticker_id].special_tag = special_tag = str(trade_tick[13])
                if special_tag != "":
                    if special_tag == 'H':
                        self.parent_app.touchline_data[ticker_id].high = ltp
                    elif special_tag == 'L':
                        self.parent_app.touchline_data[ticker_id].low = ltp
                self.parent_app.live_data[ticker_id].tick_seq = int(trade_tick[14])
                self.parent_app.live_data[ticker_id].tick_type = 1
                try:
                    self.parent_app.live_data[ticker_id].best_bid_price = float(trade_tick[15])
                    self.parent_app.live_data[ticker_id].best_bid_qty = int(trade_tick[16])
                    self.parent_app.live_data[ticker_id].best_ask_price = float(trade_tick[17])
                    self.parent_app.live_data[ticker_id].best_ask_qty = int(trade_tick[18])
                except (IndexError, ValueError, TypeError):
                    try:
                        del self.parent_app.live_data[ticker_id].best_bid_price
                        del self.parent_app.live_data[ticker_id].best_bid_qty
                        del self.parent_app.live_data[ticker_id].best_ask_price
                        del self.parent_app.live_data[ticker_id].best_ask_qty
                    except AttributeError:
                        pass
                except Exception as e:
                    self.logger.error(e)
                if self.trade_callback:
                    try:
                        self.trade_callback(ticker_id, self.parent_app.live_data[ticker_id])
                    except Exception as e:
                        self.logger.error(f'{Style.BRIGHT}{Fore.RED}Encountered error with trade_callback - {type(e)} - {e}{Style.RESET_ALL}')
        except KeyError:
            pass
        except Exception as e:
            self.logger.error(f'{Style.BRIGHT}{Fore.RED}Encountered error with tick feed - {type(e)}{Style.RESET_ALL}')

    def handle_bid_ask_data(self, bidask_tick):
        try:
            symbol = self.contract_mapping[int(bidask_tick[0])]
            for ticker_id in self.parent_app.symbol_mkt_id_map[symbol]:
                self.parent_app.live_data[ticker_id].symbol_id = int(bidask_tick[0])
                self.parent_app.live_data[ticker_id].timestamp = datetime.strptime(bidask_tick[1], '%Y-%m-%dT%H:%M:%S')
                self.parent_app.live_data[ticker_id].best_bid_price = float(bidask_tick[2])
                self.parent_app.live_data[ticker_id].best_bid_qty = int(bidask_tick[3])
                self.parent_app.live_data[ticker_id].best_ask_price = float(bidask_tick[4])
                self.parent_app.live_data[ticker_id].best_ask_qty = int(bidask_tick[5])
                self.parent_app.live_data[ticker_id].tick_type = 2
                if self.bidask_callback:
                    try:
                        self.bidask_callback(ticker_id, self.parent_app.live_data[ticker_id])
                    except Exception as e:
                        self.logger.error(f'{Style.BRIGHT}{Fore.RED}Encountered error with bidask_callback - {type(e)} - {e}{Style.RESET_ALL}')
        except KeyError:
            pass
        except Exception as e:
            self.logger.error(f'{Style.BRIGHT}{Fore.RED}Bid-ask feed error encountered ({type(e)})- {e}{Style.RESET_ALL}')

    def handle_bar_data(self, bar_data):
        try:
            symbol = self.contract_mapping[int(bar_data[0])]
            for ticker_id in self.parent_app.symbol_mkt_id_map[symbol]:
                if '+' in self.subscription_type:
                    obj_to_edit = self.parent_app.min_live_data[ticker_id]
                else:
                    obj_to_edit = self.parent_app.live_data[ticker_id]
                obj_to_edit.symbol_id = int(bar_data[0])
                obj_to_edit.timestamp = datetime.strptime(bar_data[1], '%Y-%m-%dT%H:%M:%S')
                obj_to_edit.symbol = symbol
                obj_to_edit.open = float(bar_data[2])
                obj_to_edit.high = bar_high = float(bar_data[3])
                if bar_high > obj_to_edit.day_high:
                    obj_to_edit.day_high = self.parent_app.touchline_data[ticker_id].high = bar_high
                obj_to_edit.low = bar_low = float(bar_data[4])
                if bar_low < obj_to_edit.day_low:
                    obj_to_edit.day_low = self.parent_app.touchline_data[ticker_id].low = bar_low
                obj_to_edit.close = self.parent_app.touchline_data[ticker_id].ltp = float(bar_data[5])
                obj_to_edit.volume = float(bar_data[6])
                obj_to_edit.oi = float(bar_data[7])
                if self.bar_callback:
                    try:
                        self.bar_callback(ticker_id, obj_to_edit)
                    except Exception as e:
                        self.logger.error(f'{Style.BRIGHT}{Fore.RED}Encountered error with bar_callback - {type(e)} - {e}{Style.RESET_ALL}')
        except KeyError:
            pass
        except Exception as e:
            self.logger.error(f'{Style.BRIGHT}{Fore.RED}Bar feed encountered - {e}{Style.RESET_ALL}')

    def ws_error(self, *args):
        error = args[-1]
        if any(isinstance(error, conn_error) for conn_error in [ConnectionResetError, TimeoutError]):
            self.logger.error(f"Raising WS error = {error}")
            self.last_ping_tm = self.last_pong_tm = 0

    # noinspection PyUnusedLocal
    def ws_open(self, *args):
        self.last_ping_tm = time.time()
        self.sock.ping()
        self.sock.settimeout(15)
        if self.parent_app.symbol_mkt_id_map:
            old_symbol_map = temp_symbol_map = deepcopy(self.parent_app.symbol_mkt_id_map)  # temp_symbol_map is destructed while restarting the data
            temp_symbol_map = dict(temp_symbol_map)
            self.parent_app.symbol_mkt_id_map = defaultdict(set)
            self.logger.debug(f'{list(temp_symbol_map.keys())} needs resuming here.')
            while temp_symbol_map:
                restart_symbols = list(temp_symbol_map.keys())
                restart_req_ids = [temp_symbol_map[symbol].pop() for symbol in restart_symbols]
                self.parent_app.start_live_data(restart_symbols, req_id=restart_req_ids)
                temp_symbol_map = {symbol: symbol_set for (symbol, symbol_set) in temp_symbol_map.items() if symbol_set}
            assert old_symbol_map == self.parent_app.symbol_mkt_id_map
            self.logger.info(f'All streaming symbols have been resumed.')

    # noinspection PyUnusedLocal
    def ws_close(self, *args):
        # self.logger.error('Live WebSocket Closed')
        self.sock.close() if self.sock is not None else 0
        self.sock = None
        if self.last_ping_tm == 0 == self.last_pong_tm:
            self.logger.debug('DISCONNECTION TYPE(1) FROM SERVER...')
            try:
                time.sleep(1)
                self.run_forever(ping_interval=10, ping_timeout=5)
            except Exception as e:
                self.logger.error(f'{type(e)} in reconnection => {e}')
        if self.last_ping_tm > self.last_pong_tm:
            self.logger.debug('DISCONNECTION TYPE(2) FROM SERVER...')
            try:
                time.sleep(1)
                self.run_forever(ping_interval=10, ping_timeout=5)
            except Exception as e:
                self.logger.error(f'{type(e)} in reconnection2 => {e}')
        # self.logger.error(f'CLOSE: {self.last_ping_tm} -> {self.last_pong_tm}')



from colorama import Style, Fore
from functools import wraps
from typing import Union
from copy import deepcopy
from datetime import datetime


class TouchlineData:
    def __init__(self):
        self.symbol = None
        self.symbol_id = None
        self.timestamp = None
        self.ltp = None
        self.ltq = None
        self.atp = None
        self.ttq = None
        self.open = None
        self.high = None
        self.low = None
        self.prev_close = None
        self.oi = None
        self.prev_oi = None
        self.turnover = None
        self.best_bid_price = None
        self.best_bid_qty = None
        self.best_ask_price = None
        self.best_ask_qty = None

    def __str__(self):
        return str(self.__dict__)


class TickLiveData:
    def __init__(self, symbol):
        # --Common Variables
        self.timestamp = None
        self.symbol_id = None
        self.symbol = symbol
        self.ltp = None
        self.ltq = None
        self.atp = None
        self.ttq = None
        self.day_open = None
        self.day_high = None
        self.day_low = None
        self.prev_day_close = None
        self.oi = None
        self.prev_day_oi = None
        self.turnover = None
        self.special_tag = ""
        self.tick_seq = None
        self.best_bid_price = None
        self.best_bid_qty = None
        self.best_ask_price = None
        self.best_ask_qty = None
        self.tick_type = None  # 0 -> touchline | 1 -> trade | 2 -> bidask
        # -- Calculated common
        self._change = None
        self._change_perc = None
        self._oi_change = None
        self._oi_change_perc = None
        self.populate_using_touchline = populate_touchline_data
        # -- Calculated specific
        # -- Unused
        # self.exchange = 'NSE'
        # - For level 2 and level 3 data
        # self.bids = []
        # self.asks = []

    def __eq__(self, other):
        res = True
        assert type(self) == type(other)
        try:
            if self.tick_seq != other.tick_seq\
                    or self.symbol != other.symbol\
                    or self.special_tag != other.special_tag:
                res = False
            elif self.best_bid_price != other.best_bid_price\
                    or self.best_bid_qty != other.best_bid_qty\
                    or self.best_ask_price != other.best_ask_price\
                    or self.best_ask_qty != other.best_ask_qty:
                res = False
        except AttributeError:
            pass
        return res

    def __str__(self):
        if self.special_tag == "":
            starting_formatter = ending_formatter = ""
        else:
            if self.special_tag == "H":
                starting_formatter = f"{Style.BRIGHT}{Fore.GREEN}"
                ending_formatter = f"{Style.RESET_ALL}"
            elif self.special_tag == "L":
                starting_formatter = f"{Style.BRIGHT}{Fore.RED}"
                ending_formatter = f"{Style.RESET_ALL}"
            elif self.special_tag == "O" or self.special_tag == "OHL":
                starting_formatter = f"{Style.BRIGHT}{Fore.BLUE}"
                ending_formatter = f"{Style.RESET_ALL}"
            else:
                starting_formatter = ending_formatter = ""
        op_dict = deepcopy(self.to_dict())
        return f"{starting_formatter}{str(op_dict)}{ending_formatter}"

    def to_dict(self):
        op_dict = deepcopy(self.__dict__)
        op_dict['change'] = self.change
        op_dict['change_perc'] = self.change_perc
        op_dict['oi_change'] = self.oi_change
        op_dict['oi_change_perc'] = self.oi_change_perc
        del op_dict['_change']
        del op_dict['_change_perc']
        del op_dict['_oi_change']
        del op_dict['_oi_change_perc']
        del op_dict['populate_using_touchline']
        return op_dict

    @property
    def change(self):
        try:
            self._change = self.ltp - self.prev_day_close
        except Exception as e:
            raise TDLiveCalcError(f"Encountered other change calculation error: {e} with symbol {self.symbol} with tick_type={self.tick_type} and tick_seq={self.tick_seq}")
        finally:
            return self._change

    @property
    def change_perc(self):
        try:
            self._change_perc = self.change * 100 / self.prev_day_close
        except ZeroDivisionError:
            self._change = self._change_perc = 0
        except Exception as e:
            raise TDLiveCalcError(f"Encountered other change calculation error: {e} with symbol {self.symbol} with tick_type={self.tick_type} and tick_seq={self.tick_seq}")
        finally:
            return self._change_perc

    @property
    def oi_change(self):
        try:
            self._oi_change = self.oi - self.prev_day_oi
        except Exception as e:
            raise TDLiveCalcError(f"Encountered other change calculation error: {e} with symbol {self.symbol} with tick_type={self.tick_type} and tick_seq={self.tick_seq}")
        finally:
            return self._oi_change

    @property
    def oi_change_perc(self):
        try:
            self._oi_change_perc = self._oi_change * 100 / self.prev_day_oi
        except ZeroDivisionError:
            self._oi_change = self._oi_change_perc = 0
        except Exception as e:
            raise TDLiveCalcError(f"Encountered other change calculation error: {e} with symbol {self.symbol} with tick_type={self.tick_type} and tick_seq={self.tick_seq}")
        finally:
            return self._oi_change_perc


class MinLiveData:
    def __init__(self, symbol):
        # --Common Variables
        self.timestamp = None
        self.symbol = symbol
        self.symbol_id = None
        self.day_high = None
        self.day_low = None
        self.day_open = None
        self.prev_day_close = None
        self.prev_day_oi = None
        self.oi = None
        self.ttq = None
        # --Obj specific
        self.open = None
        self.high = None
        self.low = None
        self.close = None
        self.volume = None
        # -- Calculated common
        self._change = None
        self._change_perc = None
        self._oi_change = None
        self._oi_change_perc = None
        self.populate_using_touchline = populate_touchline_data
        # -- Calculated specific
        # --Unused
        # self.exchange = 'NSE'
        # -For level 2 and level 3 data
        # self.bids = []
        # self.asks = []

    def __eq__(self, other):
        res = True
        assert type(self) == type(other)
        if self.timestamp != other.timestamp\
                or self.symbol != other.symbol:
            res = False
        return res

    def __str__(self):
        return str(deepcopy(self.to_dict()))

    def to_dict(self):
        op_dict = deepcopy(self.__dict__)
        op_dict['change'] = self.change
        op_dict['change_perc'] = self.change_perc
        op_dict['oi_change'] = self.oi_change
        op_dict['oi_change_perc'] = self.oi_change_perc
        del op_dict['_change']
        del op_dict['_change_perc']
        del op_dict['_oi_change']
        del op_dict['_oi_change_perc']
        del op_dict['populate_using_touchline']
        return op_dict

    @property  # TODO: Use decorators for exceptions
    def change(self):
        try:
            self._change = self.close - self.prev_day_close
        except Exception as e:
            raise TDLiveCalcError(f"Encountered other change calculation error: {e} with symbol {self.symbol} with timestamp={self.timestamp}")
        finally:
            return self._change

    @property
    def change_perc(self):
        try:
            self._change_perc = self.change * 100 / self.prev_day_close
        except ZeroDivisionError:
            self._change = self._change_perc = 0
        except Exception as e:
            raise TDLiveCalcError(f"Encountered other change calculation error: {e} with symbol {self.symbol} with timestamp={self.timestamp}")
        finally:
            return self._change_perc

    @property
    def oi_change(self):
        try:
            self._oi_change = self.oi - self.prev_day_oi
        except Exception as e:
            raise TDLiveCalcError(f"Encountered other change calculation error: {e} with symbol {self.symbol} with timestamp={self.timestamp}")
        finally:
            return self._oi_change

    @property
    def oi_change_perc(self):
        try:
            self._oi_change_perc = self._oi_change * 100 / self.prev_day_oi
        except ZeroDivisionError:
            self._oi_change = self._oi_change_perc = 0
        except Exception as e:
            raise TDLiveCalcError(f"Encountered other change calculation error: {e} with symbol {self.symbol} with timestamp={self.timestamp}")
        finally:
            return self._oi_change_perc


def populate_touchline_data(data_obj: Union[TickLiveData, MinLiveData], touchline_obj: TouchlineData):
    # data_obj.timestamp = datetime.now()
    data_obj.symbol = touchline_obj.symbol
    data_obj.symbol_id = touchline_obj.symbol_id
    data_obj.day_high = touchline_obj.high
    data_obj.day_low = touchline_obj.low
    data_obj.day_open = touchline_obj.open
    data_obj.prev_day_close = touchline_obj.prev_close
    data_obj.prev_day_oi = touchline_obj.prev_oi
    data_obj.oi = touchline_obj.oi
    data_obj.ttq = touchline_obj.ttq
    if type(data_obj) is TickLiveData:
        data_obj.tick_type = 0
        data_obj.ltp = touchline_obj.ltp
    else:
        data_obj.close = touchline_obj.ltp


class TDLiveCalcError(Exception):
    def __str__(self):
        return f"{Style.BRIGHT}{Fore.RED}Something's wrong with the live calculations- {self.args[0]}{Style.RESET_ALL}"


class TDHistoricDataError(Exception):
    def __str__(self):
        return f"{Style.BRIGHT}{Fore.RED}Something's wrong with the historical data- {self.args[0]}{Style.RESET_ALL}"


class TDLiveDataError(Exception):
    def __str__(self):
        return f"{Style.BRIGHT}{Fore.RED}Something's wrong with the live data- {self.args[0]}{Style.RESET_ALL}"


class TDInvalidRequestError(Exception):
    def __str__(self):
        return f"{Style.BRIGHT}{Fore.RED}Invalid request ({self.args[0]}){Style.RESET_ALL}"


def historical_decorator(func):
    @wraps(func)
    def dec_helper(obj, contract, end_time, start_time, bar_size, options=None, bidask=False):
        if not options:
            options = {}
        if bar_size.lower() == 'tick' or bar_size.lower() == 'ticks':
            bar_size = 'tick'
            options['time_format'] = '%Y-%m-%dT%H:%M:%S'  # This is the response format
            options['processor_to_call'] = {'csv': obj.hist_csv_tick_data_to_dict_list, 'json': obj.hist_json_tick_data_to_dict_list}
        elif bar_size.lower() == 'eod':
            # start_time = start_time.split('T')[0]
            # end_time = end_time.split('T')[0]
            bar_size = 'eod'
            options['time_format'] = '%Y-%m-%d'  # This is the response format
            # options['time_format'] = '%Y-%m-%dT%H:%M:%S'  # This is the response format
            options['processor_to_call'] = {'csv': obj.hist_csv_bar_data_to_dict_list, 'json': obj.hist_json_bar_data_to_dict_list}
        else:
            bar_size = bar_size.replace(' ', '')
            if bar_size[-1] == 's':
                bar_size = bar_size[:-1]
            options['time_format'] = '%Y-%m-%dT%H:%M:%S'  # This is the response format
            options['processor_to_call'] = {'csv': obj.hist_csv_bar_data_to_dict_list, 'json': obj.hist_json_bar_data_to_dict_list}
        return func(obj, contract, end_time, start_time, bar_size, options, bidask)

    return dec_helper


def access_token_decorator(func):
    @wraps(func)
    def dec_helper(obj, *args, **kwargs):
        if obj.access_token_expiry_time < datetime.now():
            obj.hist_login()
        return func(obj, *args, **kwargs)
    return dec_helper


from .base_tests import get_connection, run_live_tests, run_historical_tests
from .defaults import DEFAULT_LIVE_PORT, DEFAULT_HIST_PORT, DEFAULT_TEST_SYMBOL_LIST

from colorama import Style, Fore


def run_all_tests(user_a, pass_a, user_b, pass_b, live_port, historical_port, symbols_to_test):
    if live_port is None:
        print(f'{Style.BRIGHT}{Fore.BLUE}Setting to default live port = {DEFAULT_LIVE_PORT}{Style.RESET_ALL}')
        live_port = DEFAULT_LIVE_PORT
    else:
        print(f'{Style.BRIGHT}{Fore.BLUE}Using given live port = {live_port}{Style.RESET_ALL}')
    if historical_port is None:
        print(f'{Style.BRIGHT}{Fore.BLUE}Setting to default hist port = {DEFAULT_HIST_PORT}{Style.RESET_ALL}')
        historical_port = DEFAULT_HIST_PORT
    else:
        print(f'{Style.BRIGHT}{Fore.BLUE}Using given historical port = {historical_port}{Style.RESET_ALL}')
    if symbols_to_test is None:
        print(f'{Style.BRIGHT}{Fore.BLUE}Setting to default symbol list = {DEFAULT_TEST_SYMBOL_LIST}{Style.RESET_ALL}')
        symbols_to_test = DEFAULT_TEST_SYMBOL_LIST
    else:
        print(f'{Style.BRIGHT}{Fore.BLUE}Using given symbol list = {symbols_to_test}{Style.RESET_ALL}')

    td_obj_with_bidask = get_connection(user_a, pass_a, live_port, historical_port)
    td_obj_without_bidask = get_connection(user_b, pass_b, live_port, historical_port)
    run_historical_tests(td_obj_with_bidask, symbols=symbols_to_test)
    run_historical_tests(td_obj_without_bidask, symbols=symbols_to_test)
